@@ -9,7 +9,6 @@
 import tests
 
 from zim.formats import *
-from zim.fs import File
 from zim.notebook import Path
 from zim.parsing import link_type
 from zim.templates import Template
@@ -24,7 +23,7 @@ if not ElementTreeModule.__name__.endswith('cElementTree'):
 class TestFormatMixin(object):
 	'''Mixin for testing formats, uses data in C{tests/data/formats/}'''
 
-	reference_xml = File('tests/data/formats/parsetree.xml').read().rstrip('\n')
+	reference_xml = tests.TEST_DATA_FOLDER.file('formats/parsetree.xml').read().rstrip('\n')
 
 	reference_data = {
 		'wiki': 'wiki.txt',
@@ -56,14 +55,13 @@ class TestFormatMixin(object):
 		'''
 		name = self.format.info['name']
 		assert name in self.reference_data, 'No reference data for format "%s"' % name
-		path = 'tests/data/formats/' + self.reference_data[name]
-		text = File(path).read()
+		basename = self.reference_data[name]
+		text = tests.TEST_DATA_FOLDER.file('formats/' + basename).read()
 
 		# No absolute paths ended up in reference
-		pwd = Dir('.')
+		pwd = tests.ZIM_SRC_FOLDER
 		self.assertFalse(pwd.path in text, 'Absolute path ended up in reference')
-		if pwd.user_path is not None:
-			self.assertFalse(pwd.user_path in text, 'Absolute path ended up in reference')
+		self.assertFalse(pwd.userpath in text, 'Absolute path ended up in reference')
 
 		return text
 
@@ -74,7 +72,7 @@ class TestFormatMixin(object):
 		# Dumper
 		wanted = self.getReferenceData()
 		reftree = tests.new_parsetree_from_xml(self.reference_xml)
-		linker = StubLinker(Dir('tests/data/formats'))
+		linker = StubLinker(tests.TEST_DATA_FOLDER.folder('formats'))
 		dumper = self.format.Dumper(linker=linker)
 		result = ''.join(dumper.dump(reftree))
 		#~ print('\n' + '>'*80 + '\n' + result + '\n' + '<'*80 + '\n')
@@ -702,6 +700,15 @@ hmmm
 		t = self.format.Parser().parse([text])
 		self.assertEqual(t.tostring(), xml)
 
+	def testAnchor(self):
+		text = '{{id: test}}'
+		xml = '''\
+<?xml version='1.0' encoding='utf-8'?>
+<zim-tree><p><anchor name="test">test</anchor>
+</p></zim-tree>'''
+		tree = self.format.Parser().parse(text)
+		self.assertEqual(tree.tostring(), xml)
+
 
 class TestGFMAutolinks(tests.TestCase):
 	# See https://github.github.com/gfm/#autolinks-extension-
@@ -756,6 +763,14 @@ class TestGFMAutolinks(tests.TestCase):
 		('https://bugs.kde.org/buglist.cgi?resolution=---', True, None),
 		('https://vimhelp.org/options.txt.html#\'iskeyword\'', True, None),
 		('https://example.com/foo]', True, None),
+
+		# Zim extensions
+		('https://localhost', True, None),
+		('https://localhost/path', True, None),
+		('file:///home/foo', True, None),
+		('file://home/foo', True, None),
+		('file:/home/foo', True, None),
+		('foo://bar', True, None),
 	)
 
 	def testFunctions(self):
@@ -808,9 +823,9 @@ class TestHtmlFormat(tests.TestCase, TestFormatMixin):
 			template_options={'empty_lines': 'default'}
 		).dump(tree)
 		self.assertEqual(''.join(html),
-			'<h1>head1</h1>\n\n'
+			'<h1>head1<a id="head1" class="h_anchor"/></h1>\n\n'
 			'<br>\n\n'
-			'<h2>head2</h2>\n\n'
+			'<h2>head2<a id="head2" class="h_anchor"/></h2>\n\n'
 		)
 
 		html = self.format.Dumper(
@@ -818,8 +833,8 @@ class TestHtmlFormat(tests.TestCase, TestFormatMixin):
 			template_options={'empty_lines': 'remove'}
 		).dump(tree)
 		self.assertEqual(''.join(html),
-			'<h1>head1</h1>\n\n'
-			'<h2>head2</h2>\n\n'
+			'<h1>head1<a id="head1" class="h_anchor"/></h1>\n\n'
+			'<h2>head2<a id="head2" class="h_anchor"/></h2>\n\n'
 		)
 
 		html = self.format.Dumper(
@@ -827,8 +842,8 @@ class TestHtmlFormat(tests.TestCase, TestFormatMixin):
 			template_options={'empty_lines': 'Remove'} # case sensitive
 		).dump(tree)
 		self.assertEqual(''.join(html),
-			'<h1>head1</h1>\n\n'
-			'<h2>head2</h2>\n\n'
+			'<h1>head1<a id="head1" class="h_anchor"/></h1>\n\n'
+			'<h2>head2<a id="head2" class="h_anchor"/></h2>\n\n'
 		)
 
 
