@@ -36,6 +36,21 @@ def html_encode(text):
 		return ''
 
 
+def html_decode(text):
+	if not text is None:
+		text = text.replace('&amp;', '&')
+		text = text.replace('&lt;',  '<')
+		text = text.replace('&gt;',  '>')
+		return text
+	else:
+		return ''
+
+
+def html_to_text(html):
+	text = re.sub('<.*?>', '', html) # remove html tags
+	return html_decode(text)
+
+
 class Dumper(DumperClass):
 
 	TAGS = {
@@ -90,12 +105,13 @@ class Dumper(DumperClass):
 
 	def dump_h(self, tag, attrib, strings):
 		h = 'h' + str(attrib['level'])
+		id = heading_to_anchor(html_to_text(''.join(strings)))
 		if self._isrtl:
-			start = '<' + h + ' dir=\'rtl\'>'
+			start = '<%s dir=\'rtl\'>' % h
 		else:
-			start = '<' + h + '>'
+			start = '<%s>' % h
 		self._isrtl = None # reset
-		end = '</' + h + '>\n'
+		end = '<a id="%s" class="h_anchor"/></%s>\n' % (id, h)
 		strings.insert(0, start)
 		strings.append(end)
 		return strings
@@ -179,6 +195,10 @@ class Dumper(DumperClass):
 
 		return strings
 
+	def dump_anchor(self, tag, attrib, strings=None):
+		name = attrib['name']
+		return ['<a id="%s" class="anchor" />' % name]
+
 	def dump_link(self, tag, attrib, strings=None):
 		href = self.linker.link(attrib['href'])
 		type = link_type(attrib['href'])
@@ -187,6 +207,7 @@ class Dumper(DumperClass):
 		else:
 			text = attrib['href']
 		title = text.replace('"', '&quot;')
+		title = re.sub('<.*?>', '', title)
 		return [
 			'<a href="%s" title="%s" class="%s">%s</a>'
 				% (href, title, type, text)]
