@@ -1,5 +1,5 @@
 
-# Copyright 2008, 2012 Jaap Karssenberg <jaap.karssenberg@gmail.com>
+# Copyright 2008, 2012-2019 Jaap Karssenberg <jaap.karssenberg@gmail.com>
 
 '''This module handles parsing and dumping wiki text'''
 
@@ -8,14 +8,24 @@ import logging
 
 logger = logging.getLogger('zim.formats.wiki')
 
-from zim.parser import *
-from zim.parsing import url_re, url_encode, URL_ENCODE_DATA, \
+from zim.parser import Rule, fix_line_end, convert_space_to_tab
+from zim.parser import Parser as RuleParser
+from zim.parsing import url_encode, URL_ENCODE_DATA, \
 	escape_string, unescape_string, split_escaped_string
+from zim.parsing import url_re as old_url_re
 from zim.formats import *
 from zim.formats.plain import Dumper as TextDumper
 
+old_url_re = old_url_re.p
 
-WIKI_FORMAT_VERSION = 'zim 0.4'
+
+WIKI_FORMAT_VERSION = 'zim 0.6'
+assert WIKI_FORMAT_VERSION != 'zim 0.26' # skip number for historic reasons
+# History:
+# - 0.6  new url parsing logic
+# - 0.5  support nested formatting
+# - 0.4  parser rewrite, specifically changed parsing indented blocks
+# - 0.26 oldest format supported in python branch
 
 
 info = {
@@ -30,15 +40,16 @@ info = {
 }
 
 
-bullet_pattern = '(?:[\\*\u2022]|\\[[ \\*x>]\\]|\\d+\\.|[a-zA-Z]\\.)[\\ \\t]+'
+bullet_pattern = '(?:[\\*\u2022]|\\[[ \\*x><]\\]|\\d+\\.|[a-zA-Z]\\.)[\\ \\t]+'
 	# bullets can be '*' or 0x2022 for normal items
-	# and '[ ]', '[*]', '[x]' or '[>]' for checkbox items
+	# and '[ ]', '[*]', '[x]', '[>]' or [<]' for checkbox items
 	# and '1.', '10.', or 'a.' for numbered items (but not 'aa.')
 
 bullet_line_re = re.compile(r'^(\t*)(%s)(.*\n)$' % bullet_pattern)
 	# matches list item: prefix, bullet, text
 
 number_bullet_re = re.compile('^(\d+|[a-zA-Z])\.$')
+
 def check_number_bullet(bullet):
 	'''If bullet is a numbered bullet this returns the number or letter,
 	C{None} otherwise
@@ -66,6 +77,102 @@ def _remove_indent(text, indent):
 		# supported in python 2.6
 
 
+# NOTE: we follow rules of GFM spec, except:
+#  - we allow any URL scheme
+#  - we allow only one domain section (e.g. "localhost")
+#  - we add a file URI match
+#  - do not allow to start with "__" because of conflict with mark markup parsing
+# For GFM Markdown parser, remove these exceptions
+#
+# File paths cannot contain '\', '/', ':', '*', '?', '"', '<', '>', '|'
+# These are valid URL / path seperators: / \ : ? |
+# So restrict matching " < > and also '
+_url = r'''
+	(www\.|https?://|\w+://)			# autolink & autourl prefix
+	(?P<domain>([\w\-]+\.)*[\w\-]+)		# domain sections (GFM says "2 or more", so "+" instead of "*")
+	[^\s<]*								# any non-space char except "<"
+	'''
+_email = r'''
+	(mailto:)?
+	[\w\.\-_+]+@						# email prefix
+	([\w\-_]+\.)+[\w\-_]+				# email domain
+	'''
+_file = r'''
+	file:/+
+	[^\s"<>\']+
+	'''
+
+url_re = re.compile(
+	f'''\\b
+	(?!__)(?P<url>{_url})     |
+	(?!__)(?P<email>{_email}) |
+	(?P<fileuri>{_file})
+	''',
+	re.VERBOSE
+)
+
+url_trailing_punctuation = ('?', '!', '.', ',', ':', '*', '_', '~')
+
+
+def is_url(text):
+	'''Matches url_re and number of closing brackets matches
+	See L{https://github.github.com/gfm/#autolinks-extension-}
+	@param text: text to match as url
+	@returns: C{True} if C{text} is a valid url according to GFM rules
+	'''
+	url = match_url(text)
+	return url == text # No trailing puntuation or ")" excluded
+
+
+def match_url(text):
+	'''Match regex and count number of closing brackets
+	See L{https://github.github.com/gfm/#autolinks-extension-}
+	@param text: text to match as url
+	@returns: the url or None
+	'''
+	m = url_re.match(text)
+	if m:
+		url = m.group(0)
+		if m.lastgroup == 'email':
+			# Do not allow end in "-" or "_", use trailing "."
+			# modified rule from GFM to allow trailing __ because of mark markup
+			while url:
+				if url[-1] == '.':
+					url = url[:-1]
+				elif url[-1] == '_' and url[-2] == '_':
+					url = url[:-2]
+				elif url[-1] in ('-', '_'):
+					return None
+				else:
+					break
+			return url or None
+
+		# continue processing regular URL or file URI
+		if m.lastgroup == 'url':
+			domain = m.group('domain').split('.')
+			if '_' in domain[-1] or (len(domain) > 1 and '_' in domain[-2]):
+				# Last two domain sections cannot contain "_"
+				return None
+	else:
+		return None
+
+	while url:
+		if url[-1] in url_trailing_punctuation \
+			or (url[-1] == ')' and url.count(')') > url.count('(')):
+				url = url[:-1]
+		elif url[-1] == ';':
+			m = re.search('&\w+;$', url)
+			if m:
+				ref = m.group(0)
+				url = url[:-len(ref)]
+			else:
+				url = url[:-1]
+		else:
+			return url
+	else:
+		return None
+
+
 class WikiParser(object):
 	# This parser uses 3 levels of rules. The top level splits up
 	# paragraphs, verbatim paragraphs, images and objects.
@@ -78,10 +185,13 @@ class WikiParser(object):
 		'[x]': XCHECKED_BOX,
 		'[*]': CHECKED_BOX,
 		'[>]': MIGRATED_BOX,
+		'[<]': TRANSMIGRATED_BOX,
 		'*': BULLET,
 	}
 
-	def __init__(self):
+	def __init__(self, backward_indented_blocks=False, backward_url_parsing=False):
+		self.backward_indented_blocks = backward_indented_blocks
+		self.backward_url_parsing = backward_url_parsing
 		self.inline_parser = self._init_inline_parse()
 		self.list_and_indent_parser = self._init_intermediate_parser()
 		self.block_parser = self._init_block_parser()
@@ -93,17 +203,33 @@ class WikiParser(object):
 
 	def _init_inline_parse(self):
 		# Rules for inline formatting, links and tags
+		my_url_re = old_url_re if self.backward_url_parsing else url_re
+
+		descent = lambda *a: self.nested_inline_parser_below_link(*a)
+		self.nested_inline_parser_below_link = (
+			Rule(TAG, r'(?<!\S)@\w+', process=self.parse_tag)
+			| Rule(EMPHASIS, r'//(?!/)(.*?)(?<!:)//', descent=descent) # no ':' at the end (ex: 'http://')
+			| Rule(STRONG, r'\*\*(?!\*)(.*?)\*\*', descent=descent)
+			| Rule(MARK, r'__(?!_)(.*?)__', descent=descent)
+			| Rule(SUBSCRIPT, r'_\{(?!~)(.+?)\}', descent=descent)
+			| Rule(SUPERSCRIPT, r'\^\{(?!~)(.+?)\}', descent=descent)
+			| Rule(STRIKE, r'~~(?!~)(.+?)~~', descent=descent)
+			| Rule(VERBATIM, r"''(?!')(.+?)''")
+		)
+
+		descent = lambda *a: self.inline_parser(*a)
 		return (
-			Rule(LINK, url_re.r, process=self.parse_url) # FIXME need .r atribute because url_re is a Re object
-			| Rule(TAG, r'(?<!\S)@\w+', process=self.parse_tag)
-			| Rule(LINK, r'\[\[(?!\[)(.*?)\]\]', process=self.parse_link)
+			Rule(LINK, my_url_re, process=self.parse_url)
+			| Rule(LINK, r'\[\[(?!\[)(.*?\]*)\]\]', process=self.parse_link)
+			| Rule(ANCHOR, r'\{\{id:\s+(\w[\w-]+)\s*\}\}', process=self.parse_anchor) # HACK, hardcode inline object syntax
 			| Rule(IMAGE, r'\{\{(?!\{)(.*?)\}\}', process=self.parse_image)
-			| Rule(EMPHASIS, r'//(?!/)(.*?)//')
-			| Rule(STRONG, r'\*\*(?!\*)(.*?)\*\*')
-			| Rule(MARK, r'__(?!_)(.*?)__')
-			| Rule(SUBSCRIPT, r'_\{(?!~)(.+?)\}')
-			| Rule(SUPERSCRIPT, r'\^\{(?!~)(.+?)\}')
-			| Rule(STRIKE, r'~~(?!~)(.+?)~~')
+			| Rule(TAG, r'(?<!\S)@\w+', process=self.parse_tag)
+			| Rule(EMPHASIS, r'//(?!/)(.*?)(?<!:)//', descent=descent) # no ':' at the end (ex: 'http://')
+			| Rule(STRONG, r'\*\*(?!\*)(.*?)\*\*', descent=descent)
+			| Rule(MARK, r'__(?!_)(.*?)__', descent=descent)
+			| Rule(SUBSCRIPT, r'_\{(?!~)(.+?)\}', descent=descent)
+			| Rule(SUPERSCRIPT, r'\^\{(?!~)(.+?)\}', descent=descent)
+			| Rule(STRIKE, r'~~(?!~)(.+?)~~', descent=descent)
 			| Rule(VERBATIM, r"''(?!')(.+?)''")
 		)
 
@@ -111,7 +237,7 @@ class WikiParser(object):
 		# Intermediate level, breaks up lists and indented blocks
 		# TODO: deprecate this by taking lists out of the para
 		#       and make a new para for each indented block
-		p = Parser(
+		p = RuleParser(
 			Rule(
 				'X-Bullet-List',
 				r'''(
@@ -148,7 +274,7 @@ class WikiParser(object):
 
 	def _init_block_parser(self):
 		# Top level parser, to break up block level items
-		p = Parser(
+		p = RuleParser(
 			Rule(VERBATIM_BLOCK, r'''
 				^(?P<pre_indent>\t*) \'\'\' \s*?				# 3 "'"
 				( (?:^.*\n)*? )									# multi-line text
@@ -182,8 +308,7 @@ class WikiParser(object):
 		p.process_unmatched = self.parse_para
 		return p
 
-	@staticmethod
-	def parse_heading(builder, text):
+	def parse_heading(self, builder, text):
 		'''Parse heading and determine it's level'''
 		assert text.startswith('=')
 		for i, c in enumerate(text):
@@ -195,9 +320,11 @@ class WikiParser(object):
 			# === is level 4
 			# ...
 			# ======= is level 1
-
 		text = text[i:].lstrip() + '\n'
-		builder.append(HEADING, {'level': level}, text)
+
+		builder.start(HEADING, {'level': level})
+		self.inline_parser(builder, text)
+		builder.end(HEADING)
 
 	@staticmethod
 	def parse_pre(builder, indent, text):
@@ -342,7 +469,7 @@ class WikiParser(object):
 					pass
 				elif block.isspace():
 					builder.text(block)
-				elif self.backward \
+				elif self.backward_indented_blocks \
 				and not unindented_line_re.search(block):
 					# Before zim 0.29 all indented paragraphs were
 					# verbatim.
@@ -424,19 +551,28 @@ class WikiParser(object):
 		self.inline_parser(builder, text)
 		builder.end(BLOCK)
 
-	@staticmethod
-	def parse_link(builder, text):
+	def parse_link(self, builder, text):
 		text = text.strip('|') # old bug producing "[[|link]]", or "[[link|]]" or "[[||]]"
+		if not text or text.isspace():
+			return
+
+		href = None
 		if '|' in text:
 			href, text = text.split('|', 1)
 			text = text.strip('|') # stuff like "[[foo||bar]]"
-		else:
-			href = text
 
-		if href and not href.isspace():
-			builder.append(LINK, {'href': href}, text)
+		if text.endswith(']'):
+			delta = text.count(']') - text.count('[')
+			if delta > 0:
+				self.inline_parser.backup_parser_offset(delta)
+				text = text[:-delta]
+
+		if href is None:
+			builder.append(LINK, {'href': text}, text)
 		else:
-			pass
+			builder.start(LINK, {'href': href})
+			self.nested_inline_parser_below_link(builder, text)
+			builder.end(LINK)
 
 	@staticmethod
 	def parse_image(builder, text):
@@ -456,20 +592,34 @@ class WikiParser(object):
 		else:
 			builder.append(IMAGE, attrib)
 
-	@staticmethod
-	def parse_url(builder, text):
-		builder.append(LINK, {'href': text}, text)
+	def parse_url(self, builder, *a):
+		text = a[0]
+		if self.backward_url_parsing:
+			builder.append(LINK, {'href': text}, text)
+		else:
+			url = match_url(text)
+			if url is None:
+				self.inline_parser.backup_parser_offset(len(text) - 1)
+				builder.text(text[0]) # FIXME Ideally should allow re-parsing first character
+			elif url != text:
+				self.inline_parser.backup_parser_offset(len(text) - len(url))
+				builder.append(LINK, {'href': url}, url)
+			else:
+				builder.append(LINK, {'href': url}, url)
 
 	@staticmethod
 	def parse_tag(builder, text):
 		builder.append(TAG, {'name': text[1:]}, text)
 
+	@staticmethod
+	def parse_anchor(builder, name, *a):
+		builder.append(ANCHOR, {'name': name}, name)
+
 	@staticmethod
 	def parse_line(builder, text):
 		builder.append(LINE)
 
 
-
 wikiparser = WikiParser() #: singleton instance
 
 
@@ -477,7 +627,7 @@ wikiparser = WikiParser() #: singleton instance
 class Parser(ParserClass):
 
 	def __init__(self, version=WIKI_FORMAT_VERSION):
-		self.backward = version not in ('zim 0.26', WIKI_FORMAT_VERSION)
+		self.version = version
 
 	def parse(self, input, partial=False, file_input=False):
 		if not isinstance(input, str):
@@ -486,16 +636,22 @@ class Parser(ParserClass):
 		if not partial:
 			input = fix_line_end(input)
 
-		meta, backward = None, False
+		meta, version = None, False
 		if file_input:
 			input, meta = parse_header_lines(input)
 			version = meta.get('Wiki-Format')
-			if version and version not in ('zim 0.26', WIKI_FORMAT_VERSION):
-				backward = True
+
+		# Support backward compatibility - see history notes WIKI_FORMAT_VERSION
+		version = version or self.version
+		if version == 'zim 0.6':
+			mywikiparser = wikiparser
+		elif version in ('zim 0.4', 'zim 0.5'):
+			mywikiparser = WikiParser(backward_url_parsing=True)
+		else:
+			mywikiparser = WikiParser(backward_indented_blocks=True, backward_url_parsing=True)
 
 		builder = ParseTreeBuilder(partial=partial)
-		wikiparser.backward = backward or self.backward # HACK
-		wikiparser(builder, input)
+		mywikiparser(builder, input)
 
 		parsetree = builder.get_parsetree()
 		if meta is not None:
@@ -517,6 +673,7 @@ class Dumper(TextDumper):
 		XCHECKED_BOX: '[x]',
 		CHECKED_BOX: '[*]',
 		MIGRATED_BOX: '[>]',
+		TRANSMIGRATED_BOX: '[<]',
 		BULLET: '*',
 	}
 
@@ -563,13 +720,16 @@ class Dumper(TextDumper):
 		strings.append(' ' + tag)
 		return strings
 
+	def dump_anchor(self, tag, attrib, strings=None):
+		return ("{{id: ", attrib['name'], "}}")
+
 	def dump_link(self, tag, attrib, strings=None):
 		assert 'href' in attrib, \
 			'BUG: link misses href: %s "%s"' % (attrib, strings)
 		href = attrib['href']
 
 		if not strings or href == ''.join(strings):
-			if url_re.match(href):
+			if is_url(href):
 				return (href,) # no markup needed
 			else:
 				return ('[[', href, ']]')
@@ -591,9 +751,9 @@ class Dumper(TextDumper):
 			src += '?%s' % '&'.join(opts)
 
 		if alt:
-			return ('{{', src, '|', alt, '}}')
+			return ['{{', src, '|', alt, '}}']
 		else:
-			return('{{', src, '}}')
+			return ['{{', src, '}}']
 
 		# TODO use text for caption (with full recursion)
 
@@ -602,7 +762,7 @@ class Dumper(TextDumper):
 
 		opts = []
 		for key, value in sorted(list(attrib.items())):
-			# TODO: sorted to make order predictable for testing - prefer use of OrderedDict
+			# TODO: sorted to make order predictable for testing - prefer use of DefinitionOrderedDict
 			if key in ('type', 'indent') or value is None:
 				continue
 			# double quotes are escaped by doubling them
