@@ -166,8 +166,8 @@ URL_ENCODE_DATA = 0 # all
 URL_ENCODE_PATH = 1	# all except '/'
 URL_ENCODE_READABLE = 2 # only space and utf-8
 
-_url_encode_re = re.compile(r'[^A-Za-z0-9\-_.!~*\'()]') # unreserved
-_url_encode_path_re = re.compile(r'[^A-Za-z0-9\-_.!~*\'()/]') # unreserved + /
+_url_encode_re = re.compile(r'[^A-Za-z0-9\-_.~]') # unreserved (see rfc3986)
+_url_encode_path_re = re.compile(r'[^A-Za-z0-9\-_.~/]') # unreserved + /
 
 
 def _url_encode_on_error(error):
@@ -416,6 +416,9 @@ is_uri_re = Re('^(\w[\w\+\-\.]*):')
 	# "scheme:"
 is_url_re = Re('^(\w[\w\+\-\.]*)://')
 	# "scheme://"
+is_www_link_re = Re('^www\.([\w\-]+\.)+[\w\-]+')
+	# "www." followed by 2 or more domain sections
+	# See also 'url_re' in 'formats/wiki.py' following the GFM scheme
 is_email_re = Re('^(mailto:\S+|[^\s:]+)\@\S+\.\w+(\?.+)?$', re.U)
 	# "mailto:" address
 	# name "@" host
@@ -433,6 +436,13 @@ is_interwiki_re = Re('^(\w[\w\+\-\.]*)\?(.*)', re.U)
 is_interwiki_keyword_re = re.compile('^\w[\w+\-.]*$', re.U)
 
 
+def valid_interwiki_key(name):
+	key = re.sub('[^\w+\-.]', '_', name)
+	if key[0] in ('-', '.'):
+		key = '_' + key[1:] # "_" matches \w
+	return key
+
+
 _classes = {'c': r'[^\s"<>\']'} # limit the character class a bit
 url_re = Re(r'''(
 	\b \w[\w\+\-\.]+:// %(c)s* \[ %(c)s+ \] (?: %(c)s+ [\w/] )?  |
@@ -469,6 +479,8 @@ def link_type(link):
 			type = is_url_re[1]
 	elif is_email_re.match(link):
 		type = 'mailto'
+	elif is_www_link_re.match(link):
+		type = 'http'
 	elif '@' in link and (
 		link.startswith('mid:') or
 		link.startswith('cid:')
