@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 
 # Copyright 2009 Jaap Karssenberg <jaap.karssenberg@gmail.com>
 
@@ -75,7 +74,7 @@ KEYWORDS = (
 	'links', 'linksfrom', 'linksto', 'tag'
 )
 
-keyword_re = Re('('+'|'.join(KEYWORDS)+'):(.*)', re.I)
+keyword_re = Re('(' + '|'.join(KEYWORDS) + '):(.*)', re.I)
 operators_re = Re(r'^(\|\||\&\&|\+|\-)')
 tag_re = Re(r'^\@(\w+)$', re.U)
 
@@ -126,6 +125,7 @@ class Query(object):
 	def __init__(self, string):
 		self.string = string
 		self.root = self._parse_query(string)
+		self.find_input = self._generate_find_input()
 
 	def _parse_query(self, string):
 		# First do a raw tokenizer
@@ -142,11 +142,15 @@ class Query(object):
 				tokens.append(operators[w.lower()])
 			elif keyword_re.match(w):
 				keyword = keyword_re[1].lower()
-				string = keyword_re[2] or words.pop(0)
-				string = unescape_quoted_string(string)
-				if keyword == 'links':
-					keyword = 'linksfrom'
-				tokens.append(QueryTerm(keyword, string))
+				if not (keyword_re[2] or words):
+					# edge case - something ending in ":" but nothing following
+					tokens.append(QueryTerm('contentorname', keyword_re[1]+":")) # default keyword
+				else:
+					string = keyword_re[2] or words.pop(0)
+					string = unescape_quoted_string(string)
+					if keyword == 'links':
+						keyword = 'linksfrom'
+					tokens.append(QueryTerm(keyword, string))
 			else:
 				w = unescape_quoted_string(w)
 				if tag_re.match(w):
@@ -195,19 +199,31 @@ class Query(object):
 		#~ print root
 		return root
 
-	@property
-	def simple_match(self):
-		'''Used to determine a simple matching string to be used
-		in the find method in the pageview. Used by L{SearchDialog}
-		to set the L{PageView} find string to highligh matches in the page.
-		'''
-		# TODO make this return a list with positive terms for content
-		# if find supports an OR operator, highlight them all
-		if len(self.root) == 1 and isinstance(self.root[0], QueryTerm) \
-		and self.root[0].keyword in ('content', 'contentorname'):
-			return self.root[0].string
+	def _generate_find_input(self):
+		# parse query and format as a string or regex for the pageview "find"
+		# function - used to highlight matches in the pageview
+		strings = list(self._walk_text_content(self.root))
+		if not strings:
+			return None, None
+		elif len(strings) == 1:
+			return strings[0], False
 		else:
-			return None
+			return '|'.join(re.escape(s) for s in strings if s), True
+
+	def _walk_text_content(self, group):
+		for member in group:
+			if isinstance(member, QueryGroup):
+				for s in self._walk_text_content(member): # recurs
+					yield s
+			else: # QueryTerm
+				if member.inverse: # OPERATOR_NOT
+					pass
+				elif member.keyword in ('content', 'contentorname'):
+					yield member.string.strip('*') # strip "*" for partial matches
+				elif member.keyword == 'tag':
+					yield '@' + member.string.lstrip('@').strip('*')
+				else:
+					pass # other terms select pages, but no (easy) match in the page
 
 
 class PageSelection(set):
@@ -297,7 +313,7 @@ class SearchSelection(PageSelection):
 		results = None
 		for term in indexterms:
 			results, scope = op_func(results, scope,
-				self._process_from_index(term, scope) )
+				self._process_from_index(term, scope))
 
 		if callback:
 			if group.operator == OPERATOR_AND:
@@ -319,7 +335,7 @@ class SearchSelection(PageSelection):
 
 		for term in subgroups:
 			results, scope = op_func(results, scope,
-				self._process_group(term, scope, callbackwrapper) )
+				self._process_group(term, scope, callbackwrapper))
 
 			if callback:
 				if group.operator == OPERATOR_AND:
@@ -338,7 +354,7 @@ class SearchSelection(PageSelection):
 			myscope = scope # local copy here, need to pass full scope to _process_content
 			if term.keyword == 'contentorname':
 				results, myscope = op_func(results, myscope,
-					self._process_from_index(term, myscope, scoring=10) )
+					self._process_from_index(term, myscope, scoring=10))
 
 		if callback and (
 			group.operator == OPERATOR_OR or
@@ -399,19 +415,21 @@ class SearchSelection(PageSelection):
 				regex = self._namespace_regex(term.string)
 			elif term.keyword == 'contentorname':
 				# More lax matching for default case
-				regex = self._name_regex('*'+term.string.strip('*')+'*')
+				regex = self._name_regex('*' + term.string.strip('*') + '*')
 				term.name_regex = regex # needed in _process_content
 			else:
 				regex = self._name_regex(term.string)
 
-			#~ print '!! REGEX: ' + regex.pattern
+			#~ print('!! REGEX: ' + regex.pattern)
 			for path in generator:
 				if regex.match(path.name):
 					myresults.add(path)
 
 		elif term.keyword in ('linksfrom', 'linksto'):
-			if term.keyword == 'linksfrom': dir = LINK_DIR_FORWARD
-			else: dir = LINK_DIR_BACKWARD
+			if term.keyword == 'linksfrom':
+				dir = LINK_DIR_FORWARD
+			else:
+				dir = LINK_DIR_BACKWARD
 
 			if term.string.endswith('*'):
 				recurs = True
@@ -499,13 +517,13 @@ class SearchSelection(PageSelection):
 						pass
 			generator = page_generator()
 		else:
-			generator = map(self.notebook.get_page, self.notebook.pages.walk())
+			generator = list(map(self.notebook.get_page, self.notebook.pages.walk()))
 
 		if results is None:
 			results = SearchSelection(None)
 
 		for page in generator:
-			#~ print '!! Search content', page
+			#~ print('!! Search content', page)
 			try:
 				tree = page.get_parsetree()
 			except:
@@ -519,7 +537,7 @@ class SearchSelection(PageSelection):
 			if operator == OPERATOR_AND:
 				score = 0
 				for term in terms:
-					#~ print '!! Count AND %s' % term
+					#~ print('!! Count AND %s' % term)
 					myscore = tree.countre(term.content_regex)
 					if term.keyword == 'contentorname' \
 					and term.name_regex.match(path.name):
@@ -536,7 +554,7 @@ class SearchSelection(PageSelection):
 					self._count_score(path, score)
 			else: # OPERATOR_OR
 				for term in terms:
-					#~ print '!! Count OR %s' % term
+					#~ print('!! Count OR %s' % term)
 					score = tree.countre(term.content_regex)
 					if term.keyword == 'contentorname' \
 					and term.name_regex.match(path.name):
@@ -577,13 +595,13 @@ class SearchSelection(PageSelection):
 		regex = prefix + re.escape(string) + postfix
 
 		if case:
-			return re.compile(regex)
+			return re.compile(regex, re.U)
 		else:
-			return re.compile(regex, re.I)
+			return re.compile(regex, re.U | re.I)
 
 	def _namespace_regex(self, string, case=False):
 		# like _name_regex but adds recursive descent below the page
-		namespace = re.escape( string.strip('*:') )
+		namespace = re.escape(string.strip('*:'))
 		regex = r'^(' + namespace + '$|' + namespace + ':)'
 		if case:
 			return re.compile(regex)
@@ -600,10 +618,18 @@ class SearchSelection(PageSelection):
 		regex = r'\S*'.join(map(re.escape, parts))
 
 		# Next add word delimiters
-		if re.search(r'^[\*\w]', string, re.U): regex = r'\b' + regex
-		if re.search(r'[\*\w]$', string, re.U): regex = regex + r'\b'
+		# Avoid adding them next to non-word characters or next to chinese
+		# charaters. Chinese is treated special because it does not use
+		# whitespace as word delimiter.
+		if re.search(r'^[*\w]', string, re.U) \
+		and not '\u4e00' <= string[0] <= '\u9fff':
+			regex = r'\b' + regex
 
-		#~ print 'SEARCH REGEX: >>%s<<' % regex
+		if re.search(r'[*\w]$', string, re.U) \
+		and not '\u4e00' <= string[-1] <= '\u9fff':
+			regex = regex + r'\b'
+
+		#~ print('SEARCH REGEX: >>%s<<' % regex)
 		if case:
 			return re.compile(regex, re.U)
 		else:
